/* util/pom.c */

// Phase of the Moon.  Calculates the current phase of the moon.
// Based on routines from `Practical Astronomy with Your Calculator',
// by Duffett-Smith.  Comments give the section from the book that
// particular piece of code was adapted from.
//
// -- Keith E. Brandt  VIII 1984
//
// Rewritten for clarity by Byron Stanoszek, October 23, 1997

#include <sys/time.h>
#include <stdio.h>
#include <math.h>

#define EPOCH     473293890	// Seconds at January 1, 1985

#define EPSILONg  279.611371	// solar ecliptic long at EPOCH
#define RHOg      282.680403	// solar ecliptic long of perigee at EPOCH
#define ECCEN     0.01671542	// solar orbit eccentricity
#define lzero     18.251907	// lunar mean long at EPOCH
#define Pzero     192.917585	// lunar mean long of perigee at EPOCH
#define Nzero     55.204723	// lunar mean long of node at EPOCH

double dtor(), potm(), adj360();


void main()
{
  long seconds=time(NULL)-EPOCH;

  double today=potm((double)seconds);
  double next=potm((double)seconds+1);

  printf("The Moon is ");

  if(today >= 99.5)
    printf("Full");
  else if(today < 0.5)
    printf("New");
  else if(today >= 49.5 && today < 50.5)
    printf("%s Quarter", next > today?"First":"Last");
  else
    printf("%s %s", next > today?"Waxing":"Waning",
           today < 50?"Crescent":"Gibbous");

  printf(" (%.0f%% Full)\n", today);
}

// Return the phase of the moon
double potm(double seconds)
{
  double N, Msol, Ec, LambdaSol, l, Mm, Ev, Ac, A3, Mmprime;
  double A4, lprime, V, ldprime, D, Nm;

  N = 360 * seconds / 365.2422 / 86400;			// sec 42 #3
  N = adj360(N);
  Msol = N + EPSILONg - RHOg;				// sec 42 #4
  Msol = adj360(Msol);
  Ec = 360 / PI * ECCEN * sin(dtor(Msol));		// sec 42 #5
  LambdaSol = N + Ec + EPSILONg;			// sec 42 #6
  adj360(LambdaSol);
  l = 13.1763966 * seconds / 86400 + lzero;		// sec 61 #4
  l = adj360(l);
  Mm = l - (0.1114041 * seconds / 86400) - Pzero;	// sec 61 #5
  Mm = adj360(Mm);
  Nm = Nzero - (0.0529539 * seconds / 86400);		// sec 61 #6
  Nm = adj360(Nm);
  Ev = 1.2739 * sin(dtor(2*(l - LambdaSol) - Mm));	// sec 61 #7
  Ac = 0.1858 * sin(dtor(Msol));			// sec 61 #8
  A3 = 0.37 * sin(dtor(Msol));
  Mmprime = Mm + Ev - Ac - A3;				// sec 61 #9
  Ec = 6.2886 * sin(dtor(Mmprime));			// sec 61 #10
  A4 = 0.214 * sin(dtor(2 * Mmprime));			// sec 61 #11
  lprime = l + Ev + Ec - Ac + A4;			// sec 61 #12
  V = 0.6583 * sin(dtor(2 * (lprime - LambdaSol)));	// sec 61 #13
  ldprime = lprime + V;					// sec 61 #14
  D = ldprime - LambdaSol;				// sec 63 #2
  return(50 * (1 - cos(dtor(D))));			// sec 63 #3
}

// Convert degrees to radians
double dtor(double deg)
{
  return(deg * PI / 180);
}

// Adjust value so 0 <= deg <= 360
double adj360(double deg)
{
  while(deg < 0 || deg >= 360)
    if(deg < 0)
      deg+=360;
    else
      deg-=360;

  return deg;
}
